%{
/*
 * tcng.l - New TC configuration language
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"
#include "util.h"
#include "var.h"
#include "param.h"
#include "tree.h"
#include "device.h"
#include "field.h"

#include "y.tab.h"


static char *file_name = NULL;
static int lineno = 1;


static double si_prefix(char prefix,double f)
{
    static int print_milli_warning = 1;

    switch (prefix) {
	case 'm':
	    /* Alright, they asked for it. They'll get it. Badly. */
	    if (print_milli_warning) {
		yywarn("\"m\" means \"milli\", not \"Mega\"");
		print_milli_warning = 0;
	    }
	    /* I'm too kind ... */
	    return 1.0/f;
	case 'k':
	    return f;
	case 'M':
	    return f*f;
	case 'G':
	    return f*f*f;
	default:
	    return 1.0;
    }
}


%}

/* anchor mode, for "if" used as filter name */
%s ANCHOR

%%

qdisc				return TOK_QDISC;
ingress				return TOK_INGRESS;
class				{ BEGIN(INITIAL);
				  return TOK_CLASS; }
on				return TOK_ON;
<INITIAL>if			return TOK_IF;
<ANCHOR>if			return TOK_IF_ANCHOR;
filter				return TOK_FILTER;
element				return TOK_ELEMENT;
tunnel				return TOK_TUNNEL;
police				return TOK_POLICE;
else				return TOK_ELSE;
pass				return TOK_PASS;
reclassify			return TOK_RECLASSIFY;
drop				return TOK_DROP;
continue			return TOK_CONTINUE;

field				return TOK_FIELD;

cbq				return TOK_CBQ;
dsmark				return TOK_DSMARK;
fifo				return TOK_FIFO;
gred				return TOK_GRED;
prio				return TOK_PRIO;
red				return TOK_RED;
sfq				return TOK_SFQ;
tbf				return TOK_TBF;

fw				return TOK_FW;
route				return TOK_ROUTE;
rsvp				return TOK_RSVP;
tcindex				return TOK_TCINDEX;

ah				return TOK_AH;
allot				return TOK_ALLOT;
avpkt				return TOK_AVPKT;
bandwidth			return TOK_BANDWIDTH;
bands				return TOK_BANDS;
bounded				return TOK_BOUNDED;
burst				return TOK_BURST;
default				return TOK_DEFAULT;
default_index			return TOK_DEFAULT_INDEX;
dport				return TOK_DPORT;
dst				return TOK_DST;
ecn				return TOK_ECN;
esp				return TOK_ESP;
ewma				return TOK_EWMA;
fall_through			return TOK_FALL_THROUGH;
from				return TOK_FROM;
fromif				return TOK_FROMIF;
grio				return TOK_GRIO;
hash				return TOK_HASH;
isolated			return TOK_ISOLATED;
indices				return TOK_INDICES;
limit				return TOK_LIMIT;
mask				return TOK_MASK;
max				return TOK_MAX;
maxburst			return TOK_MAXBURST;
min				return TOK_MIN;
minburst			return TOK_MINBURST;
mpu				return TOK_MPU;
mtu				return TOK_MTU;
order				return TOK_ORDER;
pass_on				return TOK_PASS_ON;
peakrate			return TOK_PEAKRATE;
perturb				return TOK_PERTURB;
priomap				return TOK_PRIOMAP;
probability			return TOK_PROBABILITY;
ipproto				return TOK_IPPROTO;
quantum				return TOK_QUANTUM;
rate				return TOK_RATE;
set_tc_index			return TOK_SET_TC_INDEX;
shift				return TOK_SHIFT;
skip				return TOK_SKIP;
sport				return TOK_SPORT;
src				return TOK_SRC;
to				return TOK_TO;
value				return TOK_VALUE;
weight				return TOK_WEIGHT;

@				return TOK_HOST;
host				return TOK_HOST;

\$[a-zA-Z_][a-zA-Z0-9_]*	{ yylval.str = stralloc(yytext+1);
				  return VARIABLE; }
0[Xx][0-9a-fA-F]+|0[0-7]+|[0-9]+ { yylval.unum = strtol(yytext,NULL,0);
				  return UNUM; }
0[Bb][01]+			{ yylval.unum = strtol(yytext+2,NULL,2);
				  return UNUM; }
[0-9]+(\.[0-9]*)?([eE][-+]?[0-9]+)? { yylval.fnum = atof(yytext);
				  return FNUM; }

[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+	{ yylval.unum = ntohl(inet_addr(yytext));
				  if (yylval.unum == INADDR_NONE)
					yyerror("invalid IPv4 address");
				  return IPV4; }

<*>\n?[\t ]*			lineno += *yytext == '\n';
<*>^#\ [0-9]+\ \"[^"]*\"(\ [0-9]+)*\n { char *start = strchr(yytext,'"')+1;
				  *strchr(start,'"') = 0;
				  if (file_name) free(file_name);
				  file_name = stralloc(start);
				  lineno = strtol(yytext+2,NULL,0); }
<*>;[^\n]*\n			lineno++;

[mkMG]?bps			{ yylval.data.type = dt_rate;
				  yylval.data.u.fnum = si_prefix(*yytext,1000);
				  return UNIT_SUFFIX; }
[mkMG]?Bps			{ yylval.data.type = dt_rate;
				  yylval.data.u.fnum =
				    si_prefix(*yytext,1000)*8.0;
				  return UNIT_SUFFIX; }
[mkMG]?b(its?)?			{ yylval.data.type = dt_size;
				  yylval.data.u.fnum =
				    si_prefix(*yytext,1024)/8.0;
				  return UNIT_SUFFIX; }
[mkMG]?B(ytes?)?		{ yylval.data.type = dt_size;
				  yylval.data.u.fnum = si_prefix(*yytext,1024);
				  return UNIT_SUFFIX; }
[mkMG]?p(cks?)?			{ yylval.data.type = dt_psize;
				  yylval.data.u.fnum = si_prefix(*yytext,1000);
				  return UNIT_SUFFIX; }
[mkMG]?pps			{ yylval.data.type = dt_prate;
				  yylval.data.u.fnum = si_prefix(*yytext,1000);
				  return UNIT_SUFFIX; }
[mkMG]?s(ecs?)?			{ yylval.data.type = dt_time;
				  yylval.data.u.fnum = si_prefix(*yytext,1000);
				  return UNIT_SUFFIX; }

\.b				{ yylval.unum = 8;
				  return SIZE_QUAL; }
\.ns				{ yylval.unum = 16;
				  return SIZE_QUAL; }
\.hs				yyerror("host byte order access not supported");
\.nl				{ yylval.unum = 32;
				  return SIZE_QUAL; }
\.hl				yyerror("host byte order access not supported");

\"[^\"\n\t]+\"			{ *strrchr(yytext,'"') = 0;
				  yylval.str = stralloc(yytext+1);
				  return WORD; }
[a-zA-Z_][a-zA-Z0-9_]*		{ FIELD *field = field_find(yytext);
				  if (field) {
					yylval.field = field;
					return FIELD_NAME;
				  }
				  yylval.str = stralloc(yytext);
				  return WORD; }

"<<"				return SHIFT_LEFT;
">>"				return SHIFT_RIGHT;
==				return REL_EQ;
!=				return REL_NE;
"<="				return REL_LE;
">="				return REL_GE;
&&				return LOGICAL_AND;
"||"				return LOGICAL_OR;

=				{ BEGIN(ANCHOR);
				  return '='; }
"}"				{ BEGIN(INITIAL);
				  return '}'; }

<*>.				return *yytext;

%%

void vyywarnf(const char *fmt,va_list ap)
{
    fflush(stdout);
    fprintf(stderr,"%s:%d: warning: ",file_name ? file_name : "(stdin)",
      lineno);
    vfprintf(stderr,fmt,ap);
    fprintf(stderr," near \"%s\"\n",yytext);
    fflush(stderr);
}


void vyyerrorf(const char *fmt,va_list ap)
{
    fflush(stdout);
    fprintf(stderr,"%s:%d: ",file_name ? file_name : "(stdin)",lineno);
    vfprintf(stderr,fmt,ap);
    fprintf(stderr," near \"%s\"\n",yytext);
    exit(1);
}


void yywarnf(const char *fmt,...)
{
    va_list ap;

    va_start(ap,fmt);
    vyywarnf(fmt,ap);
    va_end(ap);
}


void yyerrorf(const char *fmt,...)
{
    va_list ap;

    va_start(ap,fmt);
    vyyerrorf(fmt,ap);
    va_end(ap);
}


void yywarn(const char *s)
{
    yywarnf("%s",s);
}


void yyerror(const char *s)
{
    yyerrorf("%s",s);
}


void verrorf(const char *fmt,va_list ap)
{
    fflush(stdout);
    vfprintf(stderr,fmt,ap);
    fprintf(stderr,"\n");
    exit(1);
}


void errorf(const char *fmt,...)
{
    va_list ap;

    va_start(ap,fmt);
    verrorf(fmt,ap);
    va_end(ap);
}


void warnf(const char *fmt,...)
{
    va_list ap;

    fflush(stdout);
    va_start(ap,fmt);
    fprintf(stderr,"warning: ");
    vfprintf(stderr,fmt,ap);
    fprintf(stderr,"\n");
    va_end(ap);
    fflush(stderr);
}


void warn(const char *s)
{
    warnf("%s",s);
}


void error(const char *s)
{
    errorf("%s",s);
    exit(1);
}
